#pragma once


#include <iostream>
#include <vector>
#include <fstream>
#include <functional>
//#include <stdint.h> //VS2012


//--------------------------------------Node--------------------------------------

template <uint8_t DIM> //space dimension
struct Node
{
	uint32_t id; //node id
	std::vector<double> coord; //node coordinates in DCS
	bool in_node; //true: inner node, false: node is FE(finite element) vertex


	Node(); //constructor

	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, Node<DIM>& cur_node); //input from file
};


//*******************************Functions*********************************************

//Constructor
template<uint8_t DIM>
Node<DIM>::Node()
{
	id = 0;
	coord.resize(DIM);

	for (uint8_t i = 0; i < DIM; ++i)
	{
		coord[i] = 0.0L;
	}
	in_node = false;
}


//Input from file
template<uint8_t DIM>
std::ifstream& operator>>(std::ifstream& file, Node<DIM>& cur_node)
{
	for (uint8_t j = 0; j < DIM; ++j)
	{
		file >> cur_node.coord[j];
	}
	return file;
}



//--------------------------------------FiniteElement--------------------------------------

struct FiniteElement
{
	uint32_t id; //FE id
	uint16_t material_id; //FE material id
	std::vector<uint32_t> nodes; //FE vertex ides

	
	FiniteElement(); //constructor

	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, FiniteElement<FEVertexCnt>& cur_fe); //input from file
};


//*******************************Functions*********************************************

//Constructor
FiniteElement::FiniteElement()
{
	id = 0;
	material_id = 0;
	nodes.resize(4);

	for (uint8_t i = 0; i < nodes.size(); ++i)
	{
		nodes[i] = 0;
	}
}


//Input from file
std::ifstream& operator>>(std::ifstream& file, FiniteElement& cur_fe)
{
	file >> cur_fe.material_id;
	for (uint8_t i = 0; i < cur_fe.nodes.size(); ++i)
	{
		file >> cur_fe.nodes[i];
	}

	return file;
}



//--------------------------------------BoundaryFiniteElement--------------------------------------

struct BoundaryFiniteElement
{
	uint32_t id; //FE id
	uint32_t boundary_id; //BFE boundary id
	std::vector<uint32_t> nodes; //BFE boundary vertex ides
	

	BoundaryFiniteElement(); //constructor

	//With "friend" it does not work
	//friend std::ifstream& operator>>(std::ifstream& file, BoundaryFiniteElement<BFEBoundaryVertexCnt>& cur_bfe); //input from file
};


//*******************************Functions*********************************************

//Constructor
BoundaryFiniteElement::BoundaryFiniteElement()
{
	id = 0;
	boundary_id = 0;
	nodes.resize(3);
	for (uint8_t i = 0; i < nodes.size(); ++i)
	{
		nodes[i] = 0;
	}
}


//Input from file
std::ifstream& operator>>(std::ifstream& file, BoundaryFiniteElement& cur_bfe)
{
	file >> cur_bfe.boundary_id;
	for (uint8_t i = 0; i < cur_bfe.nodes.size(); ++i)
	{
		file >> cur_bfe.nodes[i];
	}

	return file;
}


//--------------------------------------Edge--------------------------------------


struct Edge
{
	uint32_t node1_id;
	uint32_t node2_id;
	uint32_t central_node_id;


	Edge() : node1_id(0), node2_id(0), central_node_id(0) {}


	Edge(uint32_t n1_id, uint32_t n2_id) : node1_id(n1_id), node2_id(n2_id)
	{
		central_node_id = 0;
	}

	bool operator ==(const Edge& edge2) const noexcept
	{
		return (((node1_id == edge2.node1_id) && (node2_id == edge2.node2_id)) || ((node1_id == edge2.node2_id) && (node2_id == edge2.node1_id)));
	}
};



//--------------------------------------HashEdge--------------------------------------
struct HashEdge
{
	size_t operator()(const Edge& edge) const
	{
		size_t seed = 0;
		seed ^= std::hash<uint32_t>()(edge.node1_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
		seed ^= std::hash<uint32_t>()(edge.node2_id) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
	
		return seed;
	}
};